import os
import uuid
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap
from arcgis.features import FeatureLayerCollection, FeatureLayer

class VersionManager(object):
    """
    VersionManager allows users to manage the branch versioning for FeatureLayerCollection
    services. The Version Management Service is responsible for exposing the management
    capabilities necessary to support feature services that work with branch versioned
    datasets.

    See the following for more information: https://developers.arcgis.com/rest/services-reference/version-management-service.htm

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required String.  The URI to the web resource.
    ---------------     --------------------------------------------------------------------
    gis                 Required GIS. The enterprise connection to the Portal site.
    ---------------     --------------------------------------------------------------------
    flc                 Optional FeatureLayerCollection. This is the parent container that
                        the branch versioning is enabled on.
    ===============     ====================================================================

    """
    _con = None
    _flc = None
    _gis = None
    _json = None
    _versions = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis, flc=None):
        """init"""
        if isinstance(gis, GIS):
            self._gis = gis
        else:
            raise ValueError('gis must be of type GIS')
        self._con = self._gis._portal.con
        self._url = url
        if isinstance(flc, FeatureLayer):
            self._flc = flc.container
        elif flc is None or \
           isinstance(flc, FeatureLayerCollection) == False:
            furl = os.path.dirname(url) + "/FeatureServer"
            self._flc = FeatureLayerCollection(url=furl,
                                               gis=self._gis)
        else:
            self._flc = flc
    #----------------------------------------------------------------------
    def __str__(self):
        return "<VersionManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the service properties"""
        if self._properties is None:
            res = self._con.get(self._url, {'f':'json'})
            self._properties = PropertyMap(res)
        return self._properties
    #----------------------------------------------------------------------
    def create(self, name,
               permission='public',
               description=""):
        """
        Create the named version off of DEFAULT. The version is associated
        with the specified feature service. During creation, the description
        and access (default is public) may be optionally set.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Required String. The name of the version
        ---------------     --------------------------------------------------------------------
        permission          Optional String. The access permissions of the new version. The
                            default access permission is public.

                            Values: private, public, protected, or hidden
        ---------------     --------------------------------------------------------------------
        description         Optional String. The description of the new version
        ===============     ====================================================================


        :return: Boolean

        """
        params = {
            'f' : 'json',
            'versionName' : name,
            'description' : description,
            'accessPermission' : permission
        }
        url = self._url + "/create"
        res = self._con.post(url, params)
        self._versions = None
        if 'success' in res:
            return res['success']
        else:
            return res
    #----------------------------------------------------------------------
    def purge(self, version, owner):
        """
        Removes a lock from a version


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        version             Required String. The name of the version that is locked.
        ---------------     --------------------------------------------------------------------
        owner               Required String. The owner of the lock.
        ===============     ====================================================================


        :return: Boolean

        """
        url = "%s/purgeLock" % self._url
        params = {
            'f': 'json',
            'version' : version,
            'lockOwner' : owner
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return False
    #----------------------------------------------------------------------
    @property
    def locks(self):
        """
        For the specified feature service, return the locks for the
        administrator or the owner of the versions that have locks.
        """
        params = {'f' : 'json'}
        url = "%s/lockInfos" % self._url
        return self._con.post(url, params)['lockInfos']
    #----------------------------------------------------------------------
    @property
    def all(self):
        """returns all visible versions on a service"""
        if self._versions is None or \
           len(self._versions) == 0:
            url = "%s/versions" % self._url
            params = {'f':'json'}
            res = self._con.get(url, params)
            self._versions = []
            if 'versions' in res:
                for v in res['versions']:
                    guid = v['versionGuid'][1:-1]
                    vurl = "%s/versions/%s" % (self._url, guid)
                    self._versions.append(Version(url=vurl,
                                                  flc=self._flc,
                                                  gis=self._gis))
            return self._versions
        return self._versions
    #----------------------------------------------------------------------
    def search(self, owner=None, show_hidden=False):
        """
        For the specified feature service, return the info of all versions
        that the client has access to. If the client is the service owner
        (the user that published the service), all versions are accessible
        and will be returned.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        owner               Optional String. A filter the versions by the owner.
        ---------------     --------------------------------------------------------------------
        show_hidden         Optional Boolean. If False (default) hidden versions will not be
                            returned.
        ===============     ====================================================================

        :return: dict

        """
        url = "%s/versionInfos" % self._url
        params = {
            'ownerFilter' : owner,
            'includeHidden' : show_hidden,
            'f' : 'json'
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def get(self, version, mode=None):
        """
        Finds and Locations a Version by it's name

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        version             Required String. This is the name of the version to locate.
        ---------------     --------------------------------------------------------------------
        mode                Optional String. This allows users to get a version in a specific
                            state of edit or read.  If None is provided (default) the version
                            is created without entering a mode.

                            Values:

                                - edit - starts editting mode
                                - read - starts reading mode
                                - None - no mode is started.  This is default.
        ===============     ====================================================================


        """
        for v in self.versions:
            if version.lower() == v.properties['versionName'].lower():
                v.mode = mode
                return v
        return
########################################################################
class Version(object):
    """
    A `Version` represents a single branch in the version tree.

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required String.  The URI to the web resource.
    ---------------     --------------------------------------------------------------------
    gis                 Required GIS. The enterprise connection to the Portal site.
    ---------------     --------------------------------------------------------------------
    flc                 Optional FeatureLayerCollection. This is the parent container that
                        the branch versioning is enabled on.
    ---------------     --------------------------------------------------------------------
    session_guid        Optional String. If a GUID is known for specific version, a user
                        can specify it.
    ---------------     --------------------------------------------------------------------
    mode                Optional String. If a user wants to start either editing or reading
                        on creation of the Version, it can be specified here.  This is useful
                        when a user is using the `Version` with a `with` statement.

                        Allowed Values:

                            + edit - starts an edit session
                            + read - starts a read session

    ===============     ====================================================================

    """
    _flc = None
    _gis = None
    _url = None
    _guid = None
    _mode = None
    _save = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 flc,
                 gis=None,
                 session_guid=None,
                 mode=None):
        """Constructor"""
        self._mode = mode
        self._url = url
        self._save = False
        if gis is None:
            from arcgis import env
            self._gis = env.active_gis
        self._gis = gis
        self._con = self._gis._portal.con
        if session_guid is None:
            self._guid = "{%s-%s-%s-%s-%s}" % (uuid.uuid4().hex[:8],
                                               uuid.uuid4().hex[:4],
                                               uuid.uuid4().hex[:4],
                                               uuid.uuid4().hex[:4],
                                               uuid.uuid4().hex[:12])
        else:
            self._guid = session_guid
        self._flc = flc
    #----------------------------------------------------------------------
    @property
    def _parcel_fabric(self):
        """
        Provides access to a parcel fabric manager

        :returns: ParcelFabricManager
        """
        if "controllerDatasetLayers" in self._flc.properties and \
           "parcelLayerId" in self._flc.properties.controllerDatasetLayers:
            from arcgis.features._parcel import ParcelFabricManager
            url = os.path.dirname(self._flc.url) + "/ParcelFabricServer"
            return ParcelFabricManager(url=url, gis=self._gis,
                                       version=self, flc=self._flc)
        return
    #----------------------------------------------------------------------
    @property
    def utility(self):
        """provides access to the utility service manager"""

        if "controllerDatasetLayers" in self._flc.properties and \
           "utilityNetworkLayerId" in self._flc.properties.controllerDatasetLayers:
            from arcgis.features._utility import UtilityNetworkManager
            url = "%s/UtilityNetworkServer" % os.path.dirname(self._flc.url)
            return UtilityNetworkManager(url=url,
                                         version=self)
        return None
    #----------------------------------------------------------------------
    def __str__(self):
        return "<Version {name} @ {guid}>".format(name=self.properties.versionName,
                                                  guid=self.properties.versionGuid)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the service properties"""
        if self._properties is None:
            res = self._con.get(self._url, {'f':'json'})
            self._properties = PropertyMap(res)
        return self._properties
    #----------------------------------------------------------------------
    @property
    def layers(self):
        """returns the layers in the FeatureLayerCollection"""
        return self._flc.layers
    #----------------------------------------------------------------------
    @property
    def mode(self):
        """
        The `mode` allows versoin editors to start and stop edit, read, or
        view mode.

        Allowed Values:

            + edit - calls the `start_editing` method and creates a lock
            + read - calls the `start_reading` method and creates a lock
            + None - terminates all sessions and lets a user view the version information (default)


        """
        return self._mode
    #----------------------------------------------------------------------
    @mode.setter
    def mode(self, value):
        """
        The `mode` allows versoin editors to start and stop edit, read, or
        view mode.

        Allowed Values:

            + edit - calls the `start_editing` method and creates a lock
            + read - calls the `start_reading` method and creates a lock
            + None - terminates all sessions and lets a user view the version information (default)


        """
        if value != self._mode:
            if str(value).lower() == 'edit':
                if self._mode == 'read':
                    self.stop_reading()
                self._mode = 'edit'
                self.start_editing()
            elif str(value).lower() == 'read':
                if self._mode == "edit":
                    self.stop_editing(save=False)
                    self._mode = "read"
            elif value is None:
                if self._mode == 'edit':
                    self.stop_editing(save=False)
                elif self._mode == 'read':
                    self.stop_reading()
    #----------------------------------------------------------------------
    def delete(self):
        """
        Deletes the current version

        :return: Boolean

        """
        url = "%s/delete" % os.path.dirname(self._url)
        params = {
            'f' : 'json',
            'versionName' : self.properties.versionName,
            'sessionID' : self._guid
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    @property
    def save_edits(self):
        """
        Get/Set the Property to Save the Changes.

        When set to true, any edits performed on the version will be saved.
        """
        return self._save
    #----------------------------------------------------------------------
    @save_edits.setter
    def save_edits(self, value):
        """
        Get/Set the Property to Save the Changes.

        When set to true, any edits performed on the version will be saved.
        """
        if value != self._save:
            self._save = value
    #----------------------------------------------------------------------
    def start_editing(self):
        """
        Starts an edit session for the current user.

        :returns: boolean
        """
        if self._mode is None:
            self._mode = 'edit'
            params = {
            'f' : 'json',
            'sessionID' : self._guid
            }
            url = "%s/startEditing" % self._url
            res = self._con.post(url, params)
            return res['success']
        return False
    #----------------------------------------------------------------------
    def stop_editing(self, save=None):
        """
        Starts an edit session for the current user.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        save                Optional Boolean. States if the values should be saved. If the value
                            is set, the `save_edits` property will be overrided.
        ===============     ====================================================================


        :returns: boolean

        """
        if self._mode == 'edit':
            self._mode = None
            if save is None:
                save = self.save_edits
            params = {
            'f' : 'json',
            'sessionID' : self._guid,
            'saveEdits': save
            }
            url = "%s/stopEditing" % self._url
            res = self._con.post(url, params)
            return res['success']
        return False
    #----------------------------------------------------------------------
    def start_reading(self):
        """
        Start reading represents a long-term service session. When `start_reading`
        is enabled, it will prevent other users from editing or reconciling the
        version.

        :returns: Boolean

        """
        if self._mode is None:
            self._mode = 'read'
            params = {
            'f' : 'json',
            'sessionID' : self._guid
            }
            url = "%s/startReading" % self._url
            res = self._con.post(url, params)
            return res['success']
        return False
    #----------------------------------------------------------------------
    def stop_reading(self):
        """
        Stops and releases a reading session.

        :returns: Boolean

        """
        if self._mode == 'read':
            self._mode = None
            params = {
            'f' : 'json',
            'sessionID' : self._guid
            }
            url = "%s/stopReading" % self._url
            res = self._con.post(url, params)
            return res['success']
        return False
    #----------------------------------------------------------------------
    def delete_forward_edits(self, moment):
        """
        If the input moment does not match a specific moment (a moment
        corresponding to an edit operation), the call will return an error.
        The client app must correctly manage the edit session's edit
        operations moments (for example, the undo/redo stack) and not
        blindly pass in a timestamp that could mistakenly delete all the
        forward moments. Thus, the input moment must be equal to a moment
        in which an edit operation for the version was applied. The call
        will also fail if the session does not have a write lock on the
        version.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        moment              Required String. Moment representing the new tail of the version;
                            all forward moments will be trimmed.
        ===============     ====================================================================

        :return: Boolean

        """
        url = "%s/deleteForwardEdits" % self._url
        params = {
            'f' : "json",
            'sessionID' : self._guid,
            'moment' : moment
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def reconcile(self,
                  end_with_conflict=False,
                  with_post=False):
        """
        Reconcile a version against the DEFAULT version. The reconcile
        operation requires that you are the only user currently editing the
        version and the only user able to edit the version throughout the
        reconcile process until you save or post. The reconcile operation
        requires that you have full permissions to all the feature classes
        that have been modified in the version being edited. The reconcile
        operation detects differences between the branch version and the
        default version and flags these differences as conflicts. If
        conflicts exist, they should be resolved.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        end_with_conflict      Optional Boolean. Specifies if the reconcile should abort when
                               conflicts are found. The default is False
        ------------------     --------------------------------------------------------------------
        with_post              Optional Boolean. If True the with_post causes a post of the current
                               version following the reconcile.
        ==================     ====================================================================

        """
        if self._mode == 'edit':
            params = {
               'f' : 'json',
               'abortIfConflict' : end_with_conflict,
               'withPost' : with_post
            }
            url = "%s/reconcile" % self._url
            res = self._con.post(url, params)
            return res['success']
        return False
    #----------------------------------------------------------------------
    def alter(self,
              owner=None,
              version=None,
              description=None,
              permission=None):
        """
        The ```alter``` operation changes the geodatabase version's name,
        description, and access permissions.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        owner               Optional String. The new name of the owner.
        ---------------     --------------------------------------------------------------------
        version             Optional String. The new name of the version.
        ---------------     --------------------------------------------------------------------
        permission          Optional String. The new access level of the version.

                            Values: private, public, protected, or hidden
        ---------------     --------------------------------------------------------------------
        description         Optional String. The description of the new version
        ===============     ====================================================================


        :return: Boolean

        """
        url = "%s/alter" % self._url
        params = {
            'f' : 'json'
        }
        if owner or\
           version or\
           description or\
           permission:
            if owner:
                params['ownerName'] = owner
            if version:
                params['versionName'] = version
            if description:
                params['description'] = description
            if permission:
                params['accessPermission'] = permission
            res  = self._con.post(url, params)
            self._properties = None
            return res['success']
        return False
    #----------------------------------------------------------------------
    def differences(self, result_type):
        """
        The ```differences``` operation allows you to view differences between
        the current version and the default version. The two versions can
        be compared to check for the following conditions.

        - Inserts - features that are present in the current version but not the default version
        - Updates - features that have different attributes or geometry in the current version than the default version
        - Deletions - features that are present in the default version but not in the current version

        Both differences and conflicts will be returned. It is the clients
        responsibility to determine which are differences, and which are conflicts.

        :returns: dict

        """
        url = "%s/differences" % self._url
        params = {
            "f" : "json",
            "sessionID" : self._guid,
            "resultType" : result_type
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def conflicts(self):
        """
        The ```conflicts``` operation allows you to view the conflicts by layer
        and type (update-update, update-delete, delete-update) that were
        identified during the last Reconcile operation. The features that
        are in conflicts will also be returned as they existed in the branch,
        ancestor, and default versions.
        """
        params = {
            'f' : 'json',
            'sessionID' : self._guid
        }
        url = "%s/conflicts" % self._url
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def inspect(self, conflicts, inspect_all=False, set_inspected=False):
        """
        The ```inspect``` operation allows the client to annotate conflicts
        from the conflict set that was obtained during the last reconcile
        operation. Users can mark the conflicts as being inspected;
        additionally, a description or note can be associated with the
        conflict.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        conflicts           Required List.  The conflicts that are being inspected (removed)
                            from the conflict set.

                            Parameter Format:

                            [
                                {
                                  "layerId" : <layerId>,
                                  "features" : [
                                    {
                                      "objectId" : <objectId>,
                                      "note" : string
                                    }
                                  ]
                                }
                            ]

                            The objectId key is required. The note parameter is optional.

        ---------------     --------------------------------------------------------------------
        inspect_all         Optional Boolean. This parameter, if true, will mark all conflicts
                            as being inspected.
        ---------------     --------------------------------------------------------------------
        set_inspected       Optional Boolean. If True, the examined values will be set to
                            inspected. If ```inspect_all``` is True, this parameter is ignored.
        ===============     ====================================================================


        :return: Boolean


        """
        url = "%s/inspectConflicts" % self._url
        params = {
            'f' : 'json',
            'sessionId' : self._guid,
            'inspectAll' : inspect_all,
            'conflicts' : conflicts,
            'setInspected' : set_inspected
        }
        res = self._con.post(url, params)
        return res['success']
    #----------------------------------------------------------------------
    def post(self):
        """
        The Post operation allows the client to post the changes in their
        branch to the default version. The client can only post changes if
        the branch version has not been modified since the last reconcile.
        If the default version has been modified in the interim, the client
        will have to reconcile again before posting.

        :return: Boolean

        """
        if self._mode == 'edit':
            url = "%s/post" % self._url
            params = {
                "f" : "json",
                "sessionID" : self._guid
            }
            res = self._con.post(url, params)
            return res['success']
        return False
    #----------------------------------------------------------------------
    def __enter__(self):
        if self._mode == 'edit':
            self.start_editing()
        elif self.mode == 'read':
            self.start_reading()
        return self
    #----------------------------------------------------------------------
    def __exit__(self, type, value, traceback):
        if self._mode == 'edit':
            self.stop_editing(save=self.save_edits)
        elif self._mode == 'read':
            self.stop_reading()
    #----------------------------------------------------------------------
    def edit(self,
             layer,
             adds=None,
             updates=None,
             deletes=None,
             use_global_ids=False,
             rollback_on_failure=True):
        """
        The `edit` operation allows users to apply changes to the current version. The edit
        session must be in the mode of `edit` or an exception will be raised.

        =====================   ===========================================
        **Inputs**              **Description**
        ---------------------   -------------------------------------------
        layer                   Required FeatureLayer. The layer to perform
                                the edit on.
        ---------------------   -------------------------------------------
        adds                    Optional FeatureSet/List. The array of
                                features to be added.
        ---------------------   -------------------------------------------
        updates                 Optional FeatureSet/List. The array of
                                features to be updateded.
        ---------------------   -------------------------------------------
        deletes                 Optional FeatureSet/List. string of OIDs to
                                remove from service
        ---------------------   -------------------------------------------
        use_global_ids          Optional boolean. Instead of referencing
                                the default Object ID field, the service
                                will look at a GUID field to track changes.
                                This means the GUIDs will be passed instead
                                of OIDs for delete, update or add features.
        ---------------------   -------------------------------------------
        rollback_on_failure     Optional boolean. Optional parameter to
                                specify if the edits should be applied only
                                if all submitted edits succeed. If false, the
                                server will apply the edits that succeed
                                even if some of the submitted edits fail.
                                If true, the server will apply the edits
                                only if all edits succeed. The default
                                value is true.
        =====================   ===========================================

        :returns: dictionary

        """
        if self._mode == 'edit':
            return layer.edit_features(adds=adds,
                                       updates=updates,
                                       deletes=deletes,
                                       gdb_version=self.properties.versionName,
                                       use_global_ids=use_global_ids,
                                       rollback_on_failure=rollback_on_failure)
        else:
            raise Exception("Version must be in `edit` mode inorder to apply edits to this version.")
        return None



"""
These functions help you use elevation analysis
"""

import logging as _logging
import arcgis
from datetime import datetime
from arcgis.features import FeatureSet
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool

_log = _logging.getLogger(__name__)

_use_async = True


def profile(input_line_features: FeatureSet = {'exceededTransferLimit': False,
                                               'spatialReference': {'latestWkid': 3857, 'wkid': 102100},
                                               'geometryType': 'esriGeometryPolyline',
                                               'fields': [{'name': 'OID', 'type': 'esriFieldTypeOID', 'alias': 'OID'},
                                                          {'name': 'Shape_Length', 'type': 'esriFieldTypeDouble',
                                                           'alias': 'Shape_Length'}], 'displayFieldName': '',
                                               'features': []},
            profile_id_field: str = None,
            dem_resolution: str = None,
            maximum_sample_distance: float = None,
            maximum_sample_distance_units: str = """Meters""",
            gis=None) -> FeatureSet:
    """


Returns elevation profiles for the input line features.

Parameters:

   input_line_features: Input Line Features (FeatureSet). Required parameter.  The line features that will be profiled over the surface inputs.

   profile_id_field: Profile ID Field (str). Optional parameter.  A unique identifier to tie profiles to their corresponding input line features.

   dem_resolution: DEM Resolution (str). Optional parameter.  The approximate spatial resolution (cell size) of the source elevation data used for the calculation. The default is 90m.The resolution keyword is an approximation of the spatial resolution of the digital elevation model. Many elevation sources are distributed with units of arc seconds, the keyword is an approximation in meters for easier understanding.FINEST    The finest units available for the extent are used.10m    the elevation source resolution is 1/3 arc second, or approximately 10 meters.30m    the elevation source resolution is 1 arc second, or approximately 30 meters.90m    the elevation source resolution is 3 arc second, or approximately 90 meters.1000m    the elevation source resolution is 30 arc seconds, or approximately 1000 meters.
      Choice list:[' ', 'FINEST', '1000m', '10m', '30m', '90m']

   maximum_sample_distance: Maximum Sample Distance (float). Optional parameter.  The maximum sample distance along the line to sample elevation values.

   maximum_sample_distance_units: Maximum Sample Distance Units (str). Optional parameter.  The units for the Maximum Sample Distance parameter.The default is meters.Meters    The units are meters. This is the default.Kilometers    The units are kilometers.Feet    The units are feet.Yards    The units are yards.Miles    The units are miles.
      Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']

gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output_profile - Output Profile as a FeatureSet

See http://ec2-35-161-157-22.us-west-2.compute.amazonaws.com:6080/arcgis/rest/directories/arcgisoutput/Tools/Elevation_GPServer/Tools_Elevation/Profile.htm for additional help.
    """
    kwargs = locals()

    param_db = {
        "input_line_features": (FeatureSet, "InputLineFeatures"),
        "profile_id_field": (str, "ProfileIDField"),
        "dem_resolution": (str, "DEMResolution"),
        "maximum_sample_distance": (float, "MaximumSampleDistance"),
        "maximum_sample_distance_units": (str, "MaximumSampleDistanceUnits"),
        "output_profile": (FeatureSet, "Output Profile"),
    }
    return_values = [
        {"name": "output_profile", "display_name": "Output Profile", "type": FeatureSet},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.elevation.url

    return _execute_gp_tool(gis, "Profile", kwargs, param_db, return_values, _use_async, url)


def viewshed(input_points: FeatureSet = {'exceededTransferLimit': False,
                                         'spatialReference': {'latestWkid': 3857, 'wkid': 102100},
                                         'geometryType': 'esriGeometryPoint',
                                         'fields': [{'name': 'OID', 'type': 'esriFieldTypeOID', 'alias': 'OID'},
                                                    {'name': 'offseta', 'type': 'esriFieldTypeDouble',
                                                     'alias': 'offseta'},
                                                    {'name': 'offsetb', 'type': 'esriFieldTypeDouble',
                                                     'alias': 'offsetb'}], 'displayFieldName': '', 'features': []},
             maximum_distance: float = None,
             maximum_distance_units: str = """Meters""",
             dem_resolution: str = None,
             observer_height: float = None,
             observer_height_units: str = """Meters""",
             surface_offset: float = None,
             surface_offset_units: str = """Meters""",
             generalize_viewshed_polygons: bool = True,
             gis=None) -> FeatureSet:
    """


Returns polygons of visible areas for a given set of input observation points.

Parameters:

   input_points: Input Point Features (FeatureSet). Required parameter.  The point features to use as the observer locations.

   maximum_distance: Maximum Distance (float). Optional parameter.  The maximum distance to calculate the viewshed.

   maximum_distance_units: Maximum Distance Units (str). Optional parameter.  The units for the Maximum Distance parameter. The default is meters.Meters    The units are meters. This is the default.Kilometers    The units are kilometers.Feet    The units are feet.Yards    The units are yards.Miles    The units are miles.
      Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']

   dem_resolution: DEM Resolution (str). Optional parameter.  The approximate spatial resolution (cell size) of the source elevation data used for the calculation. The default is 90m.The resolution keyword is an approximation of the spatial resolution of the digital elevation model. Many elevation sources are distributed with units of arc seconds, the keyword is an approximation in meters for easier understanding.FINEST    The finest units available for the extent are used.10m    the elevation source resolution is 1/3 arc second, or approximately 10 meters.30m    the elevation source resolution is 1 arc second, or approximately 30 meters.90m    the elevation source resolution is 3 arc second, or approximately 90 meters.
      Choice list:[' ', 'FINEST', '10m', '30m', '90m']

   observer_height: Observer Height (float). Optional parameter.  The height above the surface of the observer. The default value of 1.75 meters is an average height of a person. If you are looking from an elevated location such as an observation tower or a tall building, use that height instead.

   observer_height_units: Observer Height Units (str). Optional parameter.  The units for the Observer Height parameter. The default is meters.Meters    The units are meters. This is the default.Kilometers    The units are kilometers.Feet    The units are feet.Yards    The units are yards.Miles    The units are miles.
      Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']

   surface_offset: Surface Offset (float). Optional parameter.  The height above the surface of the object you are trying to see. The default value is 0. If you are trying to see buildings or wind turbines use their height here.

   surface_offset_units: Surface Offset Units (str). Optional parameter.  The units for the Surface Offset parameter. The default is meters.Meters    The units are meters. This is the default.Kilometers    The units are kilometers.Feet    The units are feet.Yards   The units are yards.Miles    The units are miles.
      Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']

   generalize_viewshed_polygons: Generalize Viewshed Polygons (bool). Optional parameter.  Determine if the viewshed polygons are to be generalized or not.The viewshed calculation is based upon a raster elevation model which creates a result with stair-stepped edges. To create a more pleasing appearance and improve performance, the default behavior is to generalize the polygons. This generalization will not change the accuracy of the result for any location more than one half of the DEM's resolution.Checked    Generalizes the results. This is the default.Unchecked    No generalization of the output polygons will occur.

gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output_viewshed - Output Viewshed as a FeatureSet

See http://ec2-35-161-157-22.us-west-2.compute.amazonaws.com:6080/arcgis/rest/directories/arcgisoutput/Tools/Elevation_GPServer/Tools_Elevation/Viewshed.htm for additional help.
    """
    kwargs = locals()

    param_db = {
        "input_points": (FeatureSet, "InputPoints"),
        "maximum_distance": (float, "MaximumDistance"),
        "maximum_distance_units": (str, "MaximumDistanceUnits"),
        "dem_resolution": (str, "DEMResolution"),
        "observer_height": (float, "ObserverHeight"),
        "observer_height_units": (str, "ObserverHeightUnits"),
        "surface_offset": (float, "SurfaceOffset"),
        "surface_offset_units": (str, "SurfaceOffsetUnits"),
        "generalize_viewshed_polygons": (bool, "GeneralizeViewshedPolygons"),
        "output_viewshed": (FeatureSet, "Output Viewshed"),
    }
    return_values = [
        {"name": "output_viewshed", "display_name": "Output Viewshed", "type": FeatureSet},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.elevation.url

    return _execute_gp_tool(gis, "Viewshed", kwargs, param_db, return_values, _use_async, url)


def summarize_elevation(input_features: FeatureSet = {},
                        feature_id_field: str = None,
                        dem_resolution: str = None,
                        include_slope_aspect: bool = False,
                        gis=None) -> FeatureSet:
    """


Calculates summary statistics of elevation for each input feature.

Parameters:

   input_features: Input Features (FeatureSet). Required parameter.  Input point, line, or area features to summarize the elevation for.

   feature_id_field: Feature ID Field (str). Optional parameter.  The Unique ID field to use for the input features.

   dem_resolution: DEM Resolution (str). Optional parameter.  The approximate spatial resolution (cell size) of the source elevation data used for the calculation. The default is 90m.The resolution keyword is an approximation of the spatial resolution of the digital elevation model. Many elevation sources are distributed with units of arc seconds, the keyword is an approximation in meters for easier understanding.FINEST    The finest units available for the extent are used.10m    the elevation source resolution is 1/3 arc second, or approximately 10 meters.30m    the elevation source resolution is 1 arc second, or approximately 30 meters.90m    the elevation source resolution is 3 arc second, or approximately 90 meters.
      Choice list:[' ', 'FINEST', '10m', '30m', '90m']

   include_slope_aspect: Include Slope and Aspect (bool). Optional parameter.  Determines if slope and aspect for the input feature(s) will be included in the output.Checked    Slope and aspect values will be included in the output.Unchecked    Only the elevation values will be included in the output.

gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output_summary - Output Summary as a FeatureSet

See http://ec2-35-161-157-22.us-west-2.compute.amazonaws.com:6080/arcgis/rest/directories/arcgisoutput/Tools/Elevation_GPServer/Tools_Elevation/SummarizeElevation.htm for additional help.
    """
    kwargs = locals()

    param_db = {
        "input_features": (FeatureSet, "InputFeatures"),
        "feature_id_field": (str, "FeatureIDField"),
        "dem_resolution": (str, "DEMResolution"),
        "include_slope_aspect": (bool, "IncludeSlopeAspect"),
        "output_summary": (FeatureSet, "Output Summary"),
    }
    return_values = [
        {"name": "output_summary", "display_name": "Output Summary", "type": FeatureSet},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.elevation.url

    return _execute_gp_tool(gis, "SummarizeElevation", kwargs, param_db, return_values, _use_async, url)





import torch
from torch import nn
import torch.nn.functional as F
from fastai.vision.image import ImageBBox
from fastai.vision.data import ObjectCategoryList, ObjectItemList

def conv_params(in_size, out_size):
    filters = [3,2,5,4]
    strides = [1,2,3] # max_stride = 3
    pads = [0,1,2,3] # max pad
    
    if out_size == 1:
        return 1, 0, in_size
    
    for filter_size in filters:
        for pad in pads:
            for stride in strides:
                if ((out_size - 1) * stride == (in_size - filter_size) + 2 * pad):
                    return stride, pad, filter_size
    return None, None, None

class StdConv(nn.Module):
    def __init__(self, nin, nout, filter_size=3, stride=2, padding=1, drop=0.1):
        super().__init__()
        self.conv = nn.Conv2d(nin, nout, filter_size, stride=stride, padding=padding)
        self.bn = nn.BatchNorm2d(nout)
        self.drop = nn.Dropout(drop)
        
    def forward(self, x): 
        return self.drop(self.bn(F.relu(self.conv(x))))
        
def flatten_conv(x,k):
    bs,nf,gx,gy = x.size()
    x = x.permute(0,2,3,1).contiguous()
    return x.view(bs,-1,nf//k)

class OutConv(nn.Module):
    def __init__(self, k, nin, num_classes, bias):
        super().__init__()
        self.k = k
        self.oconv1 = nn.Conv2d(nin, (num_classes)*k, 3, padding=1)
        self.oconv2 = nn.Conv2d(nin, 4*k, 3, padding=1)
        self.oconv1.bias.data.zero_().add_(bias)
        
    def forward(self, x):
        return [flatten_conv(self.oconv1(x), self.k),
                flatten_conv(self.oconv2(x), self.k)]
    
class SSDHead(nn.Module):
    def __init__(self, grids, anchors_per_cell, num_classes, num_features=7, drop=0.3, bias=-4.):
        super().__init__()
        self.drop = nn.Dropout(drop)
        
        self.sconvs = nn.ModuleList([])
        self.oconvs = nn.ModuleList([])
        
        self.anc_grids = grids
        
        self._k = anchors_per_cell

        
        self.sconvs.append(StdConv(512, 256, stride=1, drop=drop))
        
        
        for i in range(len(grids)):
            
            if i == 0:
                stride, pad, filter_size = conv_params(num_features, grids[i])
            else:
                stride, pad, filter_size = conv_params(grids[i-1], grids[i])
            
            if stride is None:
                print(grids[i-1], ' --> ', grids[i])
                raise Exception('cannot create model for specified grids')
                
            self.sconvs.append(StdConv(256, 256, filter_size, stride=stride, padding=pad, drop=drop))
            self.oconvs.append(OutConv(self._k, 256, num_classes=num_classes, bias=bias))
                
    def forward(self, x):
        x = self.drop(F.relu(x))
        x = self.sconvs[0](x)
        out_classes = []
        out_bboxes = []
        for sconv, oconv in zip(self.sconvs[1:], self.oconvs):
            x = sconv(x)
            out_class, out_bbox = oconv(x)
            out_classes.append(out_class)
            out_bboxes.append(out_bbox)
            
        return [torch.cat(out_classes, dim=1),
                torch.cat(out_bboxes, dim=1)]

def one_hot_embedding(labels, num_classes):
    return torch.eye(num_classes)[labels.data.cpu()]

class BCE_Loss(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.num_classes = num_classes

    def forward(self, pred, targ):
        t = one_hot_embedding(targ, self.num_classes)
        t = torch.Tensor(t[:,1:].contiguous()).to(pred.device)
        x = pred[:,1:]
        w = self.get_weight(x,t)
        return F.binary_cross_entropy_with_logits(x, t, w, size_average=False)/(self.num_classes-1)
    
    def get_weight(self,x,t): return None

class FocalLoss(BCE_Loss):
    def get_weight(self,x,t):
        alpha,gamma = 0.25,1
        p = x.sigmoid()
        pt = p*t + (1-p)*(1-t)
        w = alpha*t + (1-alpha)*(1-t)
        w = w * (1-pt).pow(gamma)
        return w.detach()

def nms(boxes, scores, overlap=0.5, top_k=100):
    keep = scores.new(scores.size(0)).zero_().long()
    if boxes.numel() == 0: return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        keep[count] = i
        count += 1
        if idx.size(0) == 1: break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w*h
        # IoU = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        IoU = inter/union  # store result in iou
        # keep only elements with an IoU <= overlap
        idx = idx[IoU.le(overlap)]
    return keep, count

class SSDObjectCategoryList(ObjectCategoryList):
    "`ItemList` for labelled bounding boxes detected using SSD."
    def analyze_pred(self, pred, thresh=0.5, nms_overlap=0.1, ssd=None):
        # def analyze_pred(pred, anchors, grid_sizes, thresh=0.5, nms_overlap=0.1, ssd=None):
        b_clas, b_bb = pred
        a_ic = ssd._actn_to_bb(b_bb, ssd._anchors.cpu(), ssd._grid_sizes.cpu())
        conf_scores, clas_ids = b_clas[:, 1:].max(1)
        conf_scores = b_clas.t().sigmoid()

        out1, bbox_list, class_list = [], [], []

        for cl in range(1, len(conf_scores)):
            c_mask = conf_scores[cl] > thresh
            if c_mask.sum() == 0: 
                continue
            scores = conf_scores[cl][c_mask]
            l_mask = c_mask.unsqueeze(1)
            l_mask = l_mask.expand_as(a_ic)
            boxes = a_ic[l_mask].view(-1, 4) # boxes are now in range[ 0, 1]
            boxes = (boxes-0.5) * 2.0        # putting boxes in range[-1, 1]
            ids, count = nms(boxes.data, scores, nms_overlap, 50) # FIX- NMS overlap hardcoded
            ids = ids[:count]
            out1.append(scores[ids])
            bbox_list.append(boxes.data[ids])
            class_list.append(torch.tensor([cl]*count))

        if len(bbox_list) == 0:
            return None #torch.Tensor(size=(0,4)), torch.Tensor()

        return torch.cat(bbox_list, dim=0), torch.cat(class_list, dim=0) # torch.cat(out1, dim=0), 

    
    def reconstruct(self, t, x):
        if t is None: return None
        bboxes, labels = t
        if len((labels - self.pad_idx).nonzero()) == 0: return ImageBBox.create(*x.size, bboxes, labels=labels, classes=self.classes, scale=False)
        i = (labels - self.pad_idx).nonzero().min()
        bboxes,labels = bboxes[i:],labels[i:]
        return ImageBBox.create(*x.size, bboxes, labels=labels, classes=self.classes, scale=False)

    
class SSDObjectItemList(ObjectItemList):
    "`ItemList` suitable for object detection."
    _label_cls,_square_show_res = SSDObjectCategoryList,False



""" Defines the Track object.
"""

from .feature_model import FeatureModel
from ._store import *
from ._schemas import TrackSchema


class Track(FeatureModel):
    """ Represents a track feature, which describes the historical location of a worker.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        feature                Optional :class:`~arcgis.features.Feature`.
                               A feature containing the assignments attributes. Mostly intended for
                               internal usage. If supplied, other parameters are ignored.
        ------------------     --------------------------------------------------------------------
        geometry               Optional :class:`Dict`.
                               A dictionary containing the assignment geometry
        ------------------     --------------------------------------------------------------------
        accuracy               Optional :class:`Float`. The accuracy of the point
        ==================     ====================================================================
    """

    def __init__(self, project, feature=None, geometry=None, accuracy=None):
        super().__init__(project, project.tracks_layer, feature)
        self._schema = TrackSchema(project.tracks_layer)
        if not feature:
            self.accuracy = accuracy
            self.geometry = geometry

    def __str__(self):
        return "{} at {}, {}".format(self.creator, self.geometry['x'], self.geometry['y'])

    def __repr__(self):
        return "<Track {}>".format(self.object_id)

    def update(self, geometry=None, accuracy=None):
        """
            Updates the track point on the server

            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            geometry               Optional :class:`Dict`.
                                   A dictionary containing the assignment geometry
            ------------------     --------------------------------------------------------------------
            accuracy               Optional :class:`Float`. The accuracy of the point
            ==================     ====================================================================
        """
        update_track(self.project, self, geometry, accuracy)

    def delete(self):
        """Deletes the track point on the server"""
        delete_tracks(self.project, [self])

    @FeatureModel.geometry.setter
    def geometry(self, value):
        self._feature.geometry = value

    @property
    def accuracy(self):
        """ The horizontal accuracy of the location measurement, in meters."""
        return self._feature.attributes[self._schema.accuracy]

    @accuracy.setter
    def accuracy(self, value):
        self._feature.attributes[self._schema.accuracy] = value



try:
    import pandas as pd
    from ._accessor import GeoAccessor, GeoSeriesAccessor, _is_geoenabled
    from ._io.fileops import from_featureclass
    __all__ = ['GeoAccessor', 'GeoSeriesAccessor', 'from_featureclass']
except ImportError:
    pass



"""
   Converts a Layer to a Spatial DataFrame
"""
from __future__ import print_function
from __future__ import division
from .. import SpatialDataFrame
from arcgis.features.layer import FeatureLayer, Table
from ..utils import chunks
import pandas as pd
import numpy as np
import json
import warnings

_look_up_types = {
    "esriFieldTypeBlob" : "object",
    "esriFieldTypeDate" : "datetime64",
    "esriFieldTypeInteger" : "int64",
    "esriFieldTypeSmallInteger" : "int32",
    "esriFieldTypeDouble" : "float64",
    "esriFieldTypeSingle" :  "float32",
    "esriFieldTypeString" : "str",
    "esriFieldTypeGeometry" : "object",
    "esriFieldTypeOID" : "int64",
    "esriFieldTypeGlobalID" : "str",
    "esriFieldTypeRaster" : "object",
    "esriFieldTypeGUID" : "str",
    "esriFieldTypeXML" : "object"
}
#--------------------------------------------------------------------------
def from_layer(layer, **kwargs):
    """
    Converts a Feature Service Layer to a Pandas' DataFrame

    Parameters:
     :layer: FeatureLayer or Table object.  If the object is a FeatureLayer
      the function will return a Spatial DataFrame, if the object is of
      type Table, the function will return a Pandas' DataFrame

    Usage:
    >>> from arcgis.arcgisserver import Layer
    >>> from arcgis import from_layer
    >>> mylayer = Layer("https://sampleserver6.arcgisonline.com/arcgis/rest" +\
                        "/services/CommercialDamageAssessment/FeatureServer/0")
    >>> sdf = from_layer(mylayer)
    >>> print(sdf)
    """
    fields = []
    if isinstance(layer, (Table, FeatureLayer)) == False:
        raise ValueError("Invalid inputs: must be FeatureLayer or Table")
    if 'maxRecordCount' in layer.properties:
        max_records = layer.properties['maxRecordCount']
    else:
        max_records = 1000
    service_count = layer.query(return_count_only=True)
    if service_count > max_records:
        frames = []
        oid_info = layer.query(return_ids_only=True)
        for ids in chunks(oid_info['objectIds'], max_records):
            ids = [str(i) for i in ids]
            sql = "%s in (%s)" % (oid_info['objectIdFieldName'],
                                  ",".join(ids))
            frames.append(layer.query(where=sql).df)
        res = pd.concat(frames, ignore_index=True)
        res.reset_index(drop=True, inplace=True)
    else:
        res = layer.query().df
        res.reset_index(drop=True, inplace=True)
    dtypes = {}
    for field in layer.properties.fields:
        dtypes[field['name']] = _look_up_types[field['type']]
        if _look_up_types[field['type']] == 'datetime64':
            res[field['name']] = pd.to_datetime(res[field['name']]/1000, unit='s')
        del field
    return res
#----------------------------------------------------------------------
def to_layer(df,
             layer,
             update_existing=True,
             add_new=False,
             truncate=False):
    """
    Sends the Spatial DataFrame information to a published service

    :Parameters:
     :df: Spatial DataFrame object
     :layer: Feature Layer or Table Layer object
     :update_existing: boolean -
     :add_new: boolean
     :truncate: if true, all records will be deleted and the dataframe
      records will replace the service data
    Output:
     A layer object
    """
    if not isinstance(df, (SpatialDataFrame)):
        raise ValueError("df must be a SpatialDatframe")
    if not isinstance(layer, (Table, FeatureLayer)):
        raise ValueError("layer must be a FeatureLayer or Table Layer")
    if truncate:
        layer.delete_features(where='1=1')
        layer.edit_features(adds=df.to_featureset().features)
    elif update_existing:
        layer.edit_features(updates=df.to_featureset().features)
    elif add_new:
        layer.edit_features(adds=df.to_featureset().features)
    return layer



"""
Classes and objects used to manage published services.
"""
from __future__ import absolute_import
from __future__ import print_function
import os
import json
import tempfile
from .._common import BaseServer
from .parameters import Extension
from arcgis._impl.common._mixins import PropertyMap
########################################################################
class ServiceManager(BaseServer):
    """
    Helper class for managing services. This class is not created by users directly. An instance of this class,
    called 'services', is available as a property of the Server object. Users call methods on this 'services' object to
    managing services.
    """
    _currentURL = None
    _url = None
    _con = None
    _json_dict = None
    _currentFolder = None
    _folderName = None
    _folders = None
    _foldersDetail = None
    _folderDetail = None
    _webEncrypted = None
    _description = None
    _isDefault = None
    _services = None
    _json = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis,
                 initialize=False,
                 sm=None):
        """Constructor

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        url                 Required string. The administration url endpoint.
        ---------------     --------------------------------------------------------------------
        gis                 Required GIS or Server object. This handles the credential management.
        ===============     ====================================================================
        """
        if sm is not None:
            self._sm = sm
        super(ServiceManager, self).__init__(gis=gis,
                                             url=url, sm=sm)
        self._con = gis
        self._url = url
        self._currentURL = url
        self._currentFolder = '/'
        if initialize:
            self._init(gis)
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f":"json"}
        try:
            result = connection.get(path=self._currentURL,
                                    params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    @property
    def _folder(self):
        """ returns current folder """
        return self._folderName
    #----------------------------------------------------------------------
    @_folder.setter
    def _folder(self, folder):
        """gets/set the current folder"""

        if folder == "" or\
             folder == "/" or \
             folder is None:
            self._currentURL = self._url
            self._services = None
            self._description = None
            self._folderName = None
            self._webEncrypted = None
            self._init()
            self._folderName = folder
        elif folder.lower() in [f.lower() for f in self.folders]:
            self._currentURL = self._url + "/%s" % folder
            self._services = None
            self._description = None
            self._folderName = None
            self._webEncrypted = None
            self._init()
            self._folderName = folder
    #----------------------------------------------------------------------
    @property
    def folders(self):
        """ returns a list of all folders """
        if self._folders is None:
            self._init()
            self._folders = self.properties['folders']
        if "/" not in self._folders:
            self._folders.append("/")
        return self._folders
    #----------------------------------------------------------------------
    def list(self, folder=None, refresh=True):
        """
        returns a list of services in the specified folder

         ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder              Required string. The name of the folder to list services from.
        ---------------     --------------------------------------------------------------------
        refresh             Optional boolean. Default is False. If True, the list of services will be
                            requested to the server, else the list will be returned from cache.
        ===============     ====================================================================


        :return: list

        """
        if folder is None:
            folder = '/'
        if folder != self._currentFolder or \
           self._services is None or refresh:
            self._currentFolder = folder
            self._folder = folder
            return self._services_list()

        return self._services_list()
    #----------------------------------------------------------------------
    def _export_services(self, folder):
        """
        Export services allows for the backup and storage of non-hosted services.

        =================   ====================================================
        **Argument**        **Description**
        -----------------   ----------------------------------------------------
        folder              required string.  This is the path to the save folder.
                            The ArcGIS Account must have access to the location
                            to write the backup file.
        =================   ====================================================

        :returns: string to the save location.

        """
        if os.path.isdir(folder) == False:
            os.makedirs(folder)
        url = self._url + "/exportServices"
        params = {
            "f" : "json",
            "location" : folder,
            "csrfPreventToken" : self._con.token
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'location' in res:
            return res['location']
        return None
    #----------------------------------------------------------------------
    def _import_services(self, file_path):
        """
        Import services allows for the backup and storage of non-hosted services.

        =================   ====================================================
        **Argument**        **Description**
        -----------------   ----------------------------------------------------
        file_path           required string.  File path with extension
                            .agssiteservices.
        =================   ====================================================

        :returns: boolean

        """
        folder = os.path.dirname(file_path)
        if os.path.isdir(folder) == False:
            os.makedirs(folder)
        url = self._url + "/importServices"
        params = {
            "f" : "json",
            "csrfPreventToken" : self._con.token
        }
        files = {'location' : file_path}
        res = self._con.post(path=url,
                             files=files,
                             postdata=params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def _services_list(self):
        """ returns the services in the current folder """
        self._services = []
        params = {
            "f" : "json"
        }
        json_dict = self._con.get(path=self._currentURL,
                                  params=params)
        if "services" in json_dict.keys():
            for s in json_dict['services']:
                from urllib.parse import quote, quote_plus, urlparse, urljoin
                u_url = self._currentURL + "/%s.%s" % (s['serviceName'], s['type'])
                parsed = urlparse(u_url)
                u_url = "{scheme}://{netloc}{path}".format(scheme=parsed.scheme,
                                                           netloc=parsed.netloc,
                                                           path=quote(parsed.path))
                self._services.append(
                    Service(url=u_url,
                            gis=self._con)
                )
        return self._services
    #----------------------------------------------------------------------
    @property
    def _extensions(self):
        """
        This resource is a collection of all the custom server object
        extensions that have been uploaded and registered with the server.
        You can register new server object extensions using the register
        extension operation. When updating an existing extension, you need
        to use the update extension operation. If an extension is no longer
        required, you can use the unregister operation to remove the
        extension from the site.
        """
        url = self._url + "/types/extensions"
        params = {'f' : 'json'}
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def publish_sd(self,
                   sd_file,
                   folder=None):
        """
        publishes a service definition file to arcgis server

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        sd_file             Required string. File path to the .sd file
        ---------------     --------------------------------------------------------------------
        folder              Optional string. This parameter allows for the override of the
                            folder option set in the SD file.
        ===============     ====================================================================


        :return: boolean

        """
        return self._sm.publish_sd(sd_file, folder)
    #----------------------------------------------------------------------
    def _find_services(self, service_type="*"):
        """
            returns a list of a particular service type on AGS

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        service_type        Required string. Type of service to find.  The allowed types
                             are: ("GPSERVER", "GLOBESERVER", "MAPSERVER",
                             "GEOMETRYSERVER", "IMAGESERVER",
                             "SEARCHSERVER", "GEODATASERVER",
                             "GEOCODESERVER", "*").  The default is *
                             meaning find all service names.
        ===============     ====================================================================


        :return: list of service name as folder/name.type

        """
        allowed_service_types = ("GPSERVER", "GLOBESERVER", "MAPSERVER",
                                 "GEOMETRYSERVER", "IMAGESERVER",
                                 "SEARCHSERVER", "GEODATASERVER",
                                 "GEOCODESERVER", "*")
        lower_types = [l.lower() for l in service_type.split(',')]
        for v in lower_types:
            if v.upper() not in allowed_service_types:
                return {"message" : "%s is not an allowed service type." % v}
        params = {
            "f" : "json"
        }
        type_services = []
        folders = self.folders
        folders.append("")
        baseURL = self._url
        for folder in folders:
            if folder == "":
                url = baseURL
            else:
                url = baseURL + "/%s" % folder
            res = self._con.get(path=url, params=params)
            if res.has_key("services"):
                for service in res['services']:
                    if service['type'].lower() in lower_types:
                        service['URL'] = url + "/%s.%s" % (service['serviceName'],
                                                           service_type)
                        type_services.append(service)
                    del service
            del res
            del folder
        return type_services
    #----------------------------------------------------------------------
    def _examine_folder(self, folder=None):
        """
        A folder is a container for GIS services. ArcGIS Server supports a
        single level hierarchy of folders.
        By grouping services within a folder, you can conveniently set
        permissions on them as a single unit. A folder inherits the
        permissions of the root folder when it is created, but you can
        change those permissions at a later time.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder              Required string. name of folder to examine.
        ===============     ====================================================================


        :return: dict

        """
        params = {'f': 'json'}
        if folder:
            url = self._url + "/" + folder
        else:
            url = self._url
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def _can_create_service(self,
                           service,
                           options=None,
                           folder_name=None,
                           service_type=None):
        """
        Use canCreateService to determine whether a specific service can be
        created on the ArcGIS Server site.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        service             Required dict. The service configuration in JSON format. For more
                            information about the service configuration options, see
                            createService. This is an optional parameter, though either the
                            service_type or service parameter must be used.
        ---------------     --------------------------------------------------------------------
        options             optional dict. This is an optional parameter that provides additional
                            information about the service, such as whether it is a hosted
                            service.
        ---------------     --------------------------------------------------------------------
        folder_name         Optional string. This is an optional parameter to indicate the folder
                            where can_create_service will check for the service.
        ---------------     --------------------------------------------------------------------
        service_type        Optional string. The type of service that can be created. This is an
                            optional parameter, though either the service type or service
                            parameter must be used.
        ===============     ====================================================================


        :return: boolean

        """
        url = self._url + "/canCreateService"
        params = {"f" : "json",
                  'service' : service}
        if options:
            params['options'] = options
        if folder_name:
            params['folderName'] = folder_name
        if service_type:
            params['serviceType'] = service_type
        return self._con.post(path=url,
                              postdata=params)

    #----------------------------------------------------------------------
    def _add_folder_permission(self, principal, is_allowed=True, folder=None):
        """
           Assigns a new permission to a role (principal). The permission
           on a parent resource is automatically inherited by all child
           resources

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        principal           Required string. Name of role to assign/disassign accesss.
        ---------------     --------------------------------------------------------------------
        is_allowed          Optional boolean. True means grant access, False means revoke.
        ---------------     --------------------------------------------------------------------
        folder              Optional string. Name of folder to assign permissions to.
        ===============     ====================================================================


        :return: dict

        """
        if folder is not None:
            u_url = self._url + "/%s/%s" % (folder, "/permissions/add")
        else:
            u_url = self._url + "/permissions/add"
        params = {
            "f" : "json",
            "principal" : principal,
            "isAllowed" : is_allowed
        }
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _folder_permissions(self, folder_name):
        """
        Lists principals which have permissions for the folder.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder_name              Optional string. Name of folder to examine permissions.
        ===============     ====================================================================


        :return: dict

        """
        u_url = self._url + "/%s/permissions" % folder_name
        params = {
            "f" : "json",
        }
        return self._con.post(path=u_url, postdata=params)
    #----------------------------------------------------------------------
    def _clean_permissions(self, principal):
        """
        Cleans all permissions that have been assigned to a role
        (principal). This is typically used when a role is deleted.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        principal           Required string. Name of role to dis-assign all accesss.
        ===============     ====================================================================


        :return: boolean
        """
        u_url = self._url + "/permissions/clean"
        params = {
            "f" : "json",
            "principal" : principal
        }
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def create_folder(self, folder_name, description=""):
        """
        Creates a unique folder name on AGS

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder_name         Required string. Name of the new folder.
        ---------------     --------------------------------------------------------------------
        description         Optional string. Description of what the folder is.
        ===============     ====================================================================

        :return: boolean
        """
        params = {
            "f" : "json",
            "folderName" : folder_name,
            "description" : description
        }
        u_url = self._url + "/createFolder"
        res = self._con.post(path=u_url, postdata=params)
        self._init()
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def delete_folder(self, folder_name):
        """
        Removes a folder on ArcGIS Server

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder_name         Required string. Name of the folder.
        ===============     ====================================================================

        :return: boolean
        """
        params = {
            "f" : "json"
        }
        if folder_name in self.folders:
            u_url = self._url + "/%s/deleteFolder" % folder_name
            res = self._con.post(path=u_url, postdata=params)
            self._init()
            if 'status' in res:
                return res['status'] == 'success'
            return res
        else:
            return False
    #----------------------------------------------------------------------
    def _delete_service(self, name, service_type, folder=None):
        """
        Deletes a service from ArcGIS Server

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Required string. Name of the service
        ---------------     --------------------------------------------------------------------
        service_type        Required string. Name of the service type.
        ---------------     --------------------------------------------------------------------
        folder              Optional string. Location of the service on ArcGIS Server.
        ===============     ====================================================================

        :return: boolean

        """
        if folder is None:
            u_url = self._url + "/%s.%s/delete" % (name,
                                                   service_type)
        else:
            u_url = self._url + "/%s/%s.%s/delete" % (folder,
                                                      name,
                                                      service_type)
        params = {
            "f" : "json"
        }
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _service_report(self, folder=None):
        """
        Provides a report on all items in a given folder.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder              Optional string. Location of the service on ArcGIS Server.
        ===============     ====================================================================

        :return: boolean
        """
        items = ["description", "status",
                 "instances", "iteminfo",
                 "properties"]
        if folder is None:
            u_url = self._url + "/report"
        else:
            u_url = self._url + "/%s/report" % folder
        params = {
            "f" : "json",
            "parameters" : items
        }
        return self._con.get(path=u_url, params=params)
    #----------------------------------------------------------------------
    @property
    def _types(self):
        """ returns the allowed services types """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/types"
        return self._con.get(path=u_url,
                             params=params)
    #----------------------------------------------------------------------
    def _federate(self):
        """
        This operation is used when federating ArcGIS Server with Portal
        for ArcGIS. It imports any services that you have previously
        published to your ArcGIS Server site and makes them available as
        items with Portal for ArcGIS. Beginning at 10.3, services are
        imported automatically as part of the federate process.
        If the automatic import of services fails as part of the federation
        process, the following severe-level message will appear in the
        server logs:
           Failed to import GIS services as items within portal.
        If this occurs, you can manually re-run the operation to import
        your services as items in the portal. Before you do this, obtain a
        portal token and then validate ArcGIS Server is federated with
        your portal using the portal website. This is done in
        My Organization > Edit Settings > Servers.
        After you run the Federate operation, specify sharing properties to
        determine which users and groups will be able to access each service.
        """
        params = {'f' : 'json'}
        url = self._url + "/federate"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _unfederate(self):
        """
        This operation is used when unfederating ArcGIS Server with Portal
        for ArcGIS. It removes any items from the portal that represent
        services running on your federated ArcGIS Server. You typically run
        this operation in preparation for a full unfederate action. For
        example, this can be performed using
               My Organization > Edit Settings > Servers
        in the portal website or the Unregister Server operation in the
        ArcGIS REST API.
        Beginning at 10.3, services are removed automatically as part of
        the unfederate process. If the automatic removal of service items
        fails as part of the unfederate process, you can manually re-run
        the operation to remove the items from the portal.
        """
        params = {'f' : 'json'}
        url = self._url + "/unfederate"
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _unregister_extension(self, extension_filename):
        """
        Unregisters all the extensions from a previously registered server
        object extension (.SOE) file.

        ======================     ====================================================================
        **Argument**               **Description**
        ----------------------     --------------------------------------------------------------------
        extension_filename         Required string. Name of the previously registered .SOE file.
        ======================     ====================================================================

        :return: boolean

        """
        params = {
            "f" : "json",
            "extensionFilename" : extension_filename
        }
        url = self._url + "/types/extensions/unregister"
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _update_extension(self, item_id):
        """
        Updates extensions that have been previously registered with the
        server. All extensions in the new .SOE file must match with
        extensions from a previously registered .SOE file.
        Use this operation to update your implementations or extension
        configuration properties.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_id             Required string. Id of the uploaded .SOE file
        ===============     ====================================================================

        :return: boolean


        """
        params = {'f':'json',
                  'id': item_id}
        url = self._url + "/types/extensions/update"
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _rename_service(self, name, service_type,
                        new_name, folder=None):
        """
        Renames a published AGS Service

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Required string.  Old service name.
        ---------------     --------------------------------------------------------------------
        service_type        Required string. The type of service.
        ---------------     --------------------------------------------------------------------
        new_name            Required string. The new service name.
        ---------------     --------------------------------------------------------------------
        folder              Optional string. Location of where the service lives, none means
                            root folder.
        ===============     ====================================================================

        :return: boolean

        """
        params = {
            "f" : "json",
            "serviceName" : name,
            "serviceType" : service_type,
            "serviceNewName" : new_name
        }
        if folder is None:
            u_url = self._url + "/renameService"
        else:
            u_url = self._url + "/%s/renameService" % folder
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        self._init()
        return res
    #----------------------------------------------------------------------
    def create_service(self, service):
        """
        Creates a new GIS service in the folder. A service is created by
        submitting a JSON representation of the service to this operation.

        The JSON representation of a service contains the following four
        sections:
         - Service Description Properties-Common properties that are shared
          by all service types. Typically, they identify a specific service.
         - Service Framework Properties-Properties targeted towards the
          framework that hosts the GIS service. They define the life cycle
          and load balancing of the service.
         - Service Type Properties -Properties targeted towards the core
          service type as seen by the server administrator. Since these
          properties are associated with a server object, they vary across
          the service types. The Service Types section in the Help
          describes the supported properties for each service.
         - Extension Properties-Represent the extensions that are enabled
          on the service. The Extension Types section in the Help describes
          the supported out-of-the-box extensions for each service type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        service             Required dict. The service is the properties to create a service.
        ===============     ====================================================================

        :return: dict

        Output:
         dictionary status message
        """
        url = self._url + "/createService"
        params = {
            "f" : "json"
        }
        if isinstance(service, str):
            params['service'] = service
        elif isinstance(service, dict):
            params['service'] = json.dumps(service)
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _stop_services(self, services):
        """
        Stops serveral services on a single server.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        services            Required list.  A list of dictionary objects. Each dictionary object
                            is defined as:
                              folder_name - The name of the folder containing the
                                service, for example, "Planning". If the service
                                resides in the root folder, leave the folder
                                property blank ("folder_name": "").
                              serviceName - The name of the service, for example,
                                "FireHydrants".
                              type - The service type, for example, "MapServer".
                                Example:
                                [{
                                  "folder_name" : "",
                                  "serviceName" : "SampleWorldCities",
                                  "type" : "MapServer"
                                }]
        ===============     ====================================================================

        :return: boolean


        """
        url = self._url + "/stopServices"
        if isinstance(services, dict):
            services = [services]
        elif isinstance(services, (list, tuple)):
            services = list(services)
        else:
            Exception("Invalid input for parameter services")
        params = {
            "f" : "json",
            "services" : {
                "services":services
            }
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _start_services(self, services):
        """
        starts serveral services on a single server

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        services            Required list.  A list of dictionary objects. Each dictionary object
                            is defined as:
                              folder_name - The name of the folder containing the
                                service, for example, "Planning". If the service
                                resides in the root folder, leave the folder
                                property blank ("folder_name": "").
                              serviceName - The name of the service, for example,
                                "FireHydrants".
                              type - The service type, for example, "MapServer".
                                Example:
                                [{
                                  "folder_name" : "",
                                  "serviceName" : "SampleWorldCities",
                                  "type" : "MapServer"
                                }]
        ===============     ====================================================================

        :return: boolean

        """
        url = self._url + "/startServices"
        if isinstance(services, dict):
            services = [services]
        elif isinstance(services, (list, tuple)):
            services = list(services)
        else:
            Exception("Invalid input for parameter services")
        params = {
            "f" : "json",
            "services" : {
                "services":services
            }
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def _edit_folder(self, description, web_encrypted=False):
        """
        This operation allows you to change the description of an existing
        folder or change the web encrypted property.
        The web encrypted property indicates if all the services contained
        in the folder are only accessible over a secure channel (SSL). When
        setting this property to true, you also need to enable the virtual
        directory security in the security configuration.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        description         Required string. A description of the folder.
        ---------------     --------------------------------------------------------------------
        web_encrypted       Optional boolean. The boolean to indicate if the services are
                            accessible over SSL only.
        ===============     ====================================================================

        :return: boolean

        """
        url = self._url + "/editFolder"
        params = {
            "f" : "json",
            "webEncrypted" : web_encrypted,
            "description" : "%s" % description
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def exists(self, folder_name, name=None, service_type=None):
        """
        This operation allows you to check whether a folder or a service
        exists. To test if a folder exists, supply only a folder_name. To
        test if a service exists in a root folder, supply both serviceName
        and service_type with folder_name=None. To test if a service exists
        in a folder, supply all three parameters.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder_name         Required string. The folder name to check for.
        ---------------     --------------------------------------------------------------------
        name                Optional string. The service name to check for.
        ---------------     --------------------------------------------------------------------
        service_type        Optional string. A service type. Allowed values:
                             GeometryServer | ImageServer | MapServer | GeocodeServer |
                             GeoDataServer | GPServer | GlobeServer | SearchServer
        ===============     ====================================================================

        :return: boolean

        """
        if folder_name and \
           name is None and \
           service_type is None:
            for folder in self.folders:
                if folder.lower() == folder_name.lower():
                    return True
                del folder
            return False
        url = self._url + "/exists"
        params = {
            "f" : "json",
            "folderName" : folder_name,
            "serviceName" : name,
            "type" : service_type
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        elif 'exists' in res:
            return res['exists']
        return res
########################################################################
class Service(BaseServer):
    """
    Represents a GIS administrative service

    **(This should not be created by a user)**

    """
    _con = None
    _frameworkProperties = None
    _recycleInterval = None
    _instancesPerContainer = None
    _maxWaitTime = None
    _minInstancesPerNode = None
    _maxIdleTime = None
    _maxUsageTime = None
    _allowedUploadFileTypes = None
    _datasets = None
    _properties = None
    _recycleStartTime = None
    _clusterName = None
    _description = None
    _isDefault = None
    _type = None
    _serviceName = None
    _isolationLevel = None
    _capabilities = None
    _loadBalancing = None
    _configuredState = None
    _maxStartupTime = None
    _private = None
    _maxUploadFileSize = None
    _keepAliveInterval = None
    _maxInstancesPerNode = None
    _json = None
    _json_dict = None
    _interceptor = None
    _provider = None
    _portalProperties = None
    _jsonProperties = None
    _url = None
    _extensions = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 gis,
                 initialize=False,
                 **kwargs):
        """
        Constructor

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        url                 Required string. The administration URL.
        ---------------     --------------------------------------------------------------------
        gis                 Required GIS. GIS or Server object
        ---------------     --------------------------------------------------------------------
        initialize          Optional boolean. fills all the properties at object creation is
                            true.
        ===============     ====================================================================


        """
        from arcgis.gis import GIS
        if isinstance(gis, GIS):
            con = gis._con
        else:
            con = gis
        super(Service, self)

        self._service_manager = kwargs.pop('service_manager', None)
        self._url = url
        self._currentURL = url
        self._con = con
        if initialize:
            self._init(self._con)
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        """ populates server admin information """
        from .parameters import Extension
        params = {
            "f" : "json"
        }
        if connection:
            json_dict = connection.get(path=self._url,
                                       params=params)
        else:
            json_dict = self._con.get(path=self._currentURL,
                                      params=params)
        self._json = json.dumps(json_dict)
        self._json_dict = json_dict
        attributes = [attr for attr in dir(self)
                      if not attr.startswith('__') and \
                      not attr.startswith('_')]
        self._properties = PropertyMap(json_dict)
        for k, v in json_dict.items():
            if k.lower() == "extensions":
                self._extensions = []
                for ext in v:
                    self._extensions.append(Extension.fromJSON(ext))
                    del ext

            del k
            del v
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def _refresh(self):
        """refreshes the object's values by re-querying the service"""
        self._init()
    #----------------------------------------------------------------------
    def _json_properties(self):
        """returns the jsonProperties"""
        if self._jsonProperties is None:
            self._init()
        return self._jsonProperties
    #----------------------------------------------------------------------
    @property
    def extensions(self):
        """lists the extensions on a service"""
        if self._extensions is None:
            self._init()
        return self._extensions
    #----------------------------------------------------------------------
    def _modify_extensions(self,
                          extension_objects=None):
        """
        enables/disables a service extension type based on the name

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        extension_objects      Required list. A list of new extensions.
        ==================     ====================================================================


        :return: boolean

        """
        if extension_objects is None:
            extension_objects = []
        if len(extension_objects) > 0 and \
           isinstance(extension_objects[0], Extension):
            self._extensions = extension_objects
            self._json_dict['extensions'] = [x.value for x in extension_objects]
            res = self.edit(str(self))
            self._json = None
            self._init()
            return res
        return False
    #----------------------------------------------------------------------
    def _has_child_permissions_conflict(self, principal, permission):
        """
        You can invoke this operation on the resource (folder or service)
        to determine if this resource has a child resource with opposing
        permissions. This operation is typically invoked before adding a
        new permission to determine if the new addition will overwrite
        existing permissions on the child resources.
        For more information, see the section on the Continuous Inheritance
        Model.
        Since this operation basically checks if the permission to be added
        will cause a conflict with permissions on a child resource, this
        operation takes the same parameters as the Add Permission operation.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        principal           Required string. Name of the role for whom the permission is being
                            assigned.
        ---------------     --------------------------------------------------------------------
        permission          Required dict. The permission dict. The format is described below.
                            Format:
                                {
                                "isAllowed": <true|false>,
                                "constraint": ""
                                }
        ===============     ====================================================================


        :return: dict

        """
        params = {
            "f" : "json",
            "principal" : principal,
            "permission" : permission
        }
        url = self._url + "/permissions/hasChildPermissionsConflict"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def start(self):
        """ starts the specific service """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/start"
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def stop(self):
        """ stops the current service """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/stop"
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def restart(self):
        """ restarts the current service """
        self.stop()
        self.start()
        return True
    #----------------------------------------------------------------------
    def rename(self, new_name):
        """
        Renames this service to the new name

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        new_name            Required string. New name of the current service.
        ===============     ====================================================================


        :return: boolean

        """
        params = {
            "f": "json",
            "serviceName": self.properties.serviceName,
            "serviceType": self.properties.type,
            "serviceNewName": new_name
        }

        u_url = self._url[:self._url.rfind('/')] + "/renameService"

        res = self._service._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def delete(self):
        """deletes a service from arcgis server"""
        params = {
            "f" : "json",
        }
        u_url = self._url + "/delete"
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def status(self):
        """ returns the status of the service """
        params = {
            "f" : "json",
        }
        u_url = self._url + "/status"
        return self._con.get(path=u_url, params=params)
    #----------------------------------------------------------------------
    @property
    def statistics(self):
        """ returns the stats for the service """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/statistics"
        return self._con.get(path=u_url, params=params)
    #----------------------------------------------------------------------
    @property
    def _permissions(self):
        """ returns the permissions for the service """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/permissions"
        return self._con.get(path=u_url, param_dict=params)
    #----------------------------------------------------------------------
    @property
    def _iteminfo(self):
        """ returns the item information """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/iteminfo"
        return self._con.get(path=u_url, params=params)
    #----------------------------------------------------------------------
    def _register_extension(self, item_id):
        """
        Registers a new server object extension file with the server.
        Before you register the file, you need to upload the .SOE file to
        the server using the Upload Data Item operation. The item_id
        returned by the upload operation must be passed to the register
        operation.
        This operation registers all the server object extensions defined
        in the .SOE file.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_id             Required string. unique ID of the item
        ===============     ====================================================================


        :return: dict

        """
        params = {
            "id" : item_id,
            "f" : "json"
        }
        url = self._url + "/types/extensions/register"
        return self._con.post(path=url,
                              postdata=params)

    #----------------------------------------------------------------------
    def _delete_item_info(self):
        """
        Deletes the item information.
        """
        params = {
            "f" : "json"
        }
        u_url = self._url + "/iteminfo/delete"
        return self._con.get(path=u_url, params=params)
    #----------------------------------------------------------------------
    def _upload_item_info(self, folder, path):
        """
        Allows for the upload of new itemInfo files such as metadata.xml

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        folder              Required string. Folder on ArcGIS Server.
        ---------------     --------------------------------------------------------------------
        path                Required string. Full path of the file to upload.
        ===============     ====================================================================


        :return: dict


        """
        files = {}
        url = self._url + "/iteminfo/upload"
        params = {
            "f" : "json",
            "folder" : folder
        }
        files['file'] = path
        return self._con.post(path=url,
                              postdata=params,
                              files=files)
    #----------------------------------------------------------------------
    def _edit_item_info(self, json_dict):
        """
        Allows for the direct edit of the service's item's information.
        To get the current item information, pull the data by calling
        iteminfo property.  This will return the default template then pass
        this object back into the editItemInfo() as a dictionary.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        json_dict           Required dict.  Item information dictionary.
        ===============     ====================================================================


        :return: dict

        """
        url = self._url + "/iteminfo/edit"
        params = {
            "f" : "json",
            "serviceItemInfo" : json.dumps(json_dict)
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def _service_manifest(self, file_type="json"):
        """
        The service manifest resource documents the data and other
        resources that define the service origins and power the service.
        This resource will tell you underlying databases and their location
        along with other supplementary files that make up the service.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        file_type           Required string.  This value can be json or xml.  json return the
                            manifest.json file.  xml returns the manifest.xml file.
        ===============     ====================================================================


        :return: string


        """

        url = self._url + "/iteminfo/manifest/manifest.%s" % file_type
        params = {
        }
        f = self._con.get(path=url,
                          params=params,
                          out_folder=tempfile.gettempdir(),
                          file_name=os.path.basename(url))
        return open(f, 'r').read()
    #----------------------------------------------------------------------
    def _add_permission(self, principal, is_allowed=True):
        """
        Assigns a new permission to a role (principal). The permission
        on a parent resource is automatically inherited by all child resources.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        principal           Required string. The role to be assigned.
        ---------------     --------------------------------------------------------------------
        is_allowed          Optional boolean. Access of resource by boolean.
        ===============     ====================================================================


        :return: boolean

        """
        u_url = self._url + "/permissions/add"
        params = {
            "f" : "json",
            "principal" : principal,
            "isAllowed" : is_allowed
        }
        res = self._con.post(path=u_url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def edit(self, service):
        """
        To edit a service, you need to submit the complete JSON
        representation of the service, which includes the updates to the
        service properties. Editing a service causes the service to be
        restarted with updated properties.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        service             Required dict. The service JSON as a dictionary.
        ===============     ====================================================================


        :return: boolean


        """
        url = self._url + "/edit"
        params = {
            "f" : "json"
        }
        if isinstance(service, str):
            params['service'] = service
        elif isinstance(service, dict):
            params['service'] = json.dumps(service)
        res = self._con.post(path=url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
